/*
  Voltmeter
  for ATmega48 (1MHz)

  T. Nakagawa

  2010-02-13
*/


#include <avr/io.h>
#include <util/delay.h>
#include "14seglcd.h"


#define VOLT_L 7495	/* 1.1Vd̓d*1000(O) */
#define VOLT_H 28090	/* AVCC̓d*1000(O) */
#define AVERAGE (256 * 4)	/* ςƂ鐔 */


static uint16_t measure(void);
static void display(uint16_t v);


int main(void) {
  uint16_t i;
  uint32_t v;

  /* |[g(gps̓vAbv) */
  PORTB = 0xff;
  PORTC = ~_BV(PC0);
  PORTD = 0xff;

  /* ADC */
  ADMUX |= (_BV(REFS1) | _BV(REFS0));
  ADCSRA |= (_BV(ADEN) | _BV(ADPS2));

  /* LCD */
  lcd_init();

  for (; ; ) {
    /* d */
    v = 0;
    ADMUX |= _BV(REFS1);	/* ADClowWɐݒ */
    for (i = 0; i < AVERAGE; i++) {
      v += measure();
    }
    v /= AVERAGE;

    /* ʂ̕\ */
    display((uint16_t)v);
  }

  return 0;
}


/*
  d̑
*/
static uint16_t measure(void) {
  uint16_t adc;
  uint32_t tmp;

  ADCSRA |= _BV(ADSC);	/* ADCJn */
  while (bit_is_set(ADCSRA, ADSC)) ;
  adc = ADCW;
  if (adc == 1023 && bit_is_set(ADMUX, REFS1)) {	/* highWɐ؂ւ */
    ADMUX &= ~_BV(REFS1);
    return measure();
  }
  tmp = adc;
  if (bit_is_set(ADMUX, REFS1)) {
    tmp *= VOLT_L;
  } else {
    tmp *= VOLT_H;
  }
  tmp /= 1024;

  return (uint16_t)tmp;
}


/*
  LCDւ̕\
*/
static void display(uint16_t v) {
  uint16_t tmp;

  lcd_buf[7] = LCD_V;
  lcd_buf[6] = LCD_SPACE;
  tmp = v % 10;
  lcd_buf[5] = (uint8_t)tmp;
  v /= 10;
  tmp = v % 10;
  lcd_buf[4] = (uint8_t)tmp;
  v /= 10;
  tmp = v % 10;
  lcd_buf[3] = (uint8_t)tmp;
  lcd_buf[2] = LCD_PERIOD;
  v /= 10;
  tmp = v % 10;
  lcd_buf[1] = (uint8_t)tmp;
  v /= 10;
  tmp = v % 10;
  lcd_buf[0] = (tmp != 0) ? (uint8_t)tmp : LCD_SPACE;

  lcd_refresh();

  return;
}
